// arbitrary precision reverse Polish calculator

const int MAXVAL_A=100;
// #define MAXFLOATPRECISIONBITS 100000

// functions
double arbitraryprecisionreversepolishcalculator(char formula[]);
void push_a(const mpf_t t);
void pop_a(mpf_t t);
int getop_a(char s[], char formula[]);
int MAXFLOATPRECISIONBITS;

// custom allocation functions, uncomment if allocation problems arise
// void *(* alloc_func_ptr ) (size_t);
// void *(* realloc_func_ptr ) (void *, size_t, size_t);
// void (*freefunc) (void *, size_t);

// variables
int pos_a;
int sp_a; /* next free stack position */
mpf_t val_a[MAXVAL_A]; /* mpf_t value stack */

double arbitraryprecisionreversepolishcalculator(char formula[])
{
int type, i;
long int iop3, iop4, iop5;
double ad;
mp_exp_t *exponent;
mpf_t op1_a, op2_a, op3_a, opt_a;
char s[MAXOP];

   pos_a=sp_a=0;

//    mp_get_memory_functions (&alloc_func_ptr, &realloc_func_ptr, &freefunc);

    // initialize arbitrary floats
    mpf_init2(op1_a, MAXFLOATPRECISIONBITS);
    mpf_init2(op2_a, MAXFLOATPRECISIONBITS);
    mpf_init2(op3_a, MAXFLOATPRECISIONBITS);
    mpf_init2(opt_a, MAXFLOATPRECISIONBITS);
    for (i=0;i<MAXVAL_A;i++)
     mpf_init2(val_a[i], MAXFLOATPRECISIONBITS);
    
    pos_a=0;
    while ((type = getop_a(s, formula)) != EOF) {
     switch (type) {
      case NUMBER:
       mpf_set_str(op1_a, s, 10);
       push_a(op1_a);
      break;
      case '+':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_add(op1_a, op2_a, op3_a);
       push_a(op1_a);
      break;
      case '*':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_mul(op1_a, op2_a, op3_a);
       push_a(op1_a);
      break;
      case '^':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_pow_ui(op1_a, op2_a, mpf_get_si(op3_a));
       push_a(op1_a);
      break;
      case '-':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_sub(op1_a, op3_a, op2_a);
       push_a(op1_a);
      break;
      case '/':
       pop_a(op2_a);
       pop_a(op3_a);
       if (!mpf_sgn(op2_a))
        mpf_set_d(op1_a, 0.0);
       else
        mpf_div(op1_a, op3_a, op2_a);
       push_a(op1_a);
      break;
      case '!':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_abs(op1_a, op3_a);
       push_a(op1_a);
      break;
      case '?':
       pop_a(op2_a);
       pop_a(op3_a);
       mpf_sqrt(op1_a, op3_a);
       push_a(op1_a);
      break;
      case '&':
       pop_a(op2_a);
       pop_a(op3_a); // op2
       mpf_urandomb(op2_a, rnd_state, 10240);
       mpf_mul(op1_a, op2_a, op3_a);
       push_a(op1_a);
      break;
      case '%':
       pop_a(op2_a);
       pop_a(op3_a);
       iop4=mpf_get_si(op2_a);
       iop5=mpf_get_si(op3_a);
       if (iop4) {
        iop3=iop5 % iop4;
       mpf_set_ui(op1_a, iop3); }
       else
        mpf_set_ui(op1_a, 0);
       push_a(op1_a);
      break;
      case '[':
       pop_a(op2_a);
       pop_a(op3_a);
       iop3=sin(mpf_get_si(op3_a));
       mpf_set_ui(op1_a, iop3);
       push_a(op1_a);
      break;
      case ']':
       pop_a(op2_a);
       pop_a(op3_a);
       iop3=cos(mpf_get_si(op3_a));
       mpf_set_ui(op1_a, iop3);
       push_a(op1_a);
      break;
      case ';':
       pop_a(op2_a);
       pop_a(op3_a);
       iop3=tan(mpf_get_si(op3_a));
       mpf_set_ui(op1_a, iop3);
       push_a(op1_a);
      break;   
      case '|':
       pop_a(op2_a);
       pop_a(op3_a);
       if (!mpf_sgn(op3_a))
        break;
       iop3=1/(tan(mpf_get_si(op3_a)));
       mpf_set_ui(op1_a, iop3);
       push_a(op1_a);
      break;
      case ':':
       pop_a(op2_a);
       pop_a(op3_a);
       iop3=log(mpf_get_si(op3_a));
       mpf_set_ui(op1_a, iop3);
       push_a(op1_a);
      break;
      case '=':
       pop_a(op2_a);
       pop_a(op3_a);
       i=mpf_cmp(op2_a, op3_a);
       if ( i == 0 )
        mpf_set_ui(opt_a, 1);
       else
        mpf_set_ui(opt_a, 0);
       push_a(opt_a);
      break;
      case '>':
       pop_a(op2_a);
       pop_a(op3_a);
       i=mpf_cmp(op2_a, op3_a);
       if ( i == -1 )
        mpf_set_ui(opt_a, 1);
       else
        mpf_set_ui(opt_a, 0);
       push_a(opt_a);
      break;
      case '}':
       pop_a(op2_a);
       pop_a(op3_a);
       i=mpf_cmp(op2_a, op3_a);
       if ( i == 0 || i == -1 )
        mpf_set_ui(opt_a, 1);
       else
        mpf_set_ui(opt_a, 0);
       push_a(opt_a);
      break;
      case '<':
       pop_a(op2_a);
       pop_a(op3_a);
       i=mpf_cmp(op2_a, op3_a);
       if ( i == 1 )
        mpf_set_ui(opt_a, 1);
       else
        mpf_set_ui(opt_a, 0);
       push_a(opt_a);
      break;
      case '{':
       pop_a(op2_a);
       pop_a(op3_a);
       i=mpf_cmp(op2_a, op3_a);
       if ( i == 0 || i == 1 )
        mpf_set_ui(opt_a, 1);
       else
        mpf_set_ui(opt_a, 0);
       push_a(opt_a);
      break;
      case '\0':
       pop_a(op1_a);
       if (!askvariable)
        gmp_printf("\t%.*Ff\n", precision_a, op1_a);
       ad=mpf_get_d(op1_a);
       // clear arbitrary floats
       mpf_clear(op1_a);
       mpf_clear(op2_a);
       mpf_clear(op3_a);
       for (i=0;i<MAXVAL_A;i++)
        mpf_clear(val_a[i]);
       // return converted to double from op1_a
       return ad;
      break;
      default:
      break;
     }
    }
    
  return 0;
}

/* push: push mpf_t onto value stack */
void push_a(const mpf_t t)
{
  if (sp_a<MAXVAL_A)
   mpf_set(val_a[sp_a++], t);
}

// pop: pop mpf_t from value stack
void pop_a(mpf_t t)
{
  if (sp_a>0)
   mpf_set(t, val_a[--sp_a]);
  else
   mpf_set_d(t, 0.0);
}

/* getop: get next character or numeric operand */
int getop_a(char s[], char formula[])
{
int i, c;

  if (pos_a>strlen(formula))
   return EOF;
  while ((s[0] = c = formula[pos_a++]) == ' ' || c == '\t');
  s[1] = '\0';
  
  /* trick getop for <= >= */
  if ( c == '<' )
   if ( formula[pos] == '=' )
    c=formula[pos++]='{';
  if ( c == '>' )
   if ( formula[pos] == '=' )
    c=formula[pos++]='}';
  
  if ( c == '_'  && isdigit2(formula[pos_a]) ) {
   c=formula[pos_a];
   s[0]='-';
  }
  
  if (!isdigit2(c) && !isdecimalseparator(c))
   return c; /* not a number */
  i = 0;
  if (isdigit2(c)) /* collect integer part */
  while (isdigit2(s[++i] = c = formula[pos_a++]));
   if (c == '.') { /* collect fraction part */
  while (isdigit2(s[++i] = c = formula[pos_a++])); }
  s[i] = '\0';

return NUMBER;
}
