// calculator interpeter
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <pwd.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <string>
#include <gmp.h>

using namespace std;

const double version=7.4;
int precision=16, precision_a=256, defineflag, askvariable=0, arbitrary=0, highlightedtext=0;
double d=0; // global variable with final formula result
gmp_randstate_t rnd_state;

#include "ansiscreen.cc"
#include "polishcalculator.cc"
#include "arbitraryprecisionreversepolishcalculator.cc"
#include "libcalint.cc"

int main(int argc, char *argv[])
{
  int i, n, formulaerrors, srcfileexecution=0, askcommand=0;
  char c, formula[MAXCOMMAND], tformula[MAXCOMMAND], ttformula[MAXCOMMAND], askedvariable[MAXNAME+1], *extract, myhomepath[MAXCOMMAND];

  struct passwd *pw = getpwuid(getuid());
  sprintf( myhomepath, "%s/.calint", pw->pw_dir);
  strcpy(formula, ".");

  // initial source file
  if ( argc > 1 && tryfile(argv[1]) )
   strcpy( srcfilename, argv[1] );
  else
   strcpy(srcfilename, "calint.ini");
  if ( tryfile(srcfilename) == 0 )
   strcpy( srcfilename, myhomepath );
  if ( (srcfileexecution=tryfile(srcfilename)) )
   newsourcefile(srcfilename);
  
  // initialize random numbers
  srand(time(NULL));
  gmp_randinit_default(rnd_state);

  initiateinternalfunctions();
  cout.precision(2);
  cout << "calint (calculator interpeter) " << showpoint << version << noshowpoint << endl;
  cout.precision(precision);
  
   // loop until exit
   while ( true ) {
       
    i=defineflag=0; emptystring[0]='\0';
    // redefine last calculated value variable
    strcpy(tformula, "lastcalculatedvalue");
    removevariable(tformula);
    definevariable(tformula, emptystring, d);
    // redefine last given command variable
    if (strcmp(ttformula, ".")) {
     strcpy(tformula, "lastgivencommand$");
     removevariable(tformula);
    definevariable(tformula, ttformula, 0); }
    // redefine, if necessary, MAXFLOATPRECISIONBITS
    strcpy(tformula, "floatprecisionbits");
    if ( !recallvariable(tformula) )
     definevariable(tformula, emptystring, 100000);
    MAXFLOATPRECISIONBITS=(int) recallvariable(tformula);
    // read .src file
    if ( srcfileexecution && !askcommand )
     srcfileexecution=executesourcefile(formula);
     
    if ( srcfileexecution == 0 || askcommand ) {
     while ( (c=getchar()) != '\n' )
      formula[i++]=c;
     formula[i]='\0';
     if (!i)
      defineflag=1;
     askcommand=0;
    }
    // skip remarks line
    if (isremline(formula)) {
     defineflag=1;
    memset(formula, 0, sizeof(formula)); }
    // recall last command, if requested
    if (!strcmp(formula, ".") && strlen(ttformula))
     strcpy(formula, ttformula);
    // keep original copy of formula
    if (strlen(formula))
     strcpy(ttformula, formula);
    // remove unnecessary spaces
    removedoublespaces(formula);
    // run aliases parser
    if (strcmp(words[separatewords(formula, 0)].formulaWord, "unalias") && strcmp(words[separatewords(formula, 0)].formulaWord, "alias"))
     aliasesreplacer(formula);
    // parse for extremely long numbers in formula
    if ((formulaerrors=parseformulaforerrors(formula, 2))) {
     defineflag=1;
     if (formulaerrors==-2) {
      cout << (char) 34 << formula << (char) 34 << " contains extremely long number(s)" << endl; 
   memset(formula, 0, sizeof(formula)); } }

   // commands
   if (!defineflag) {
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "if")) {
     defineflag=1;
     if (words.size()<4 /*|| !srcfileexecution*/)
      cout << "if condition{=<>}condition then result { else ... } is used in script files" << endl;
     else {
      i=readifconditionandresult(formula);
      if (i==-1)
       cout << (char) 34 << formula << (char) 34 << " is in wrong syntax" << endl;
      if (i)
    defineflag=0; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "evaluate")) {
     defineflag=1;
     if (words.size()!=2)
      cout << "evaluate condition{=<>}value will return true or false" << endl;
     else {
      strcpy(tformula, "if ");
      strcat(tformula, words[separatewords(formula, 1)].formulaWord);
      strcat(tformula, " then");
      i=readifconditionandresult(tformula);
      if (i==-1)
       cout << (char) 34 << formula << (char) 34 << " is in wrong syntax" << endl;
      else
    cout << "condition is " << booleans[i] << endl; } }
    // parse formula for variable recalculation
    i=parseformulaforvariablerecalculation(formula);
    if (i)
     defineflag=1;
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "exit")) {
     strcpy( tformula, "exitcalint$" );
     if ( isvariablevalid(tformula) ) {
      searchandsubstitutevariables(tformula);
      cout << tformula << endl;
     }
     break;
    }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "askcommand")) {
     defineflag=1;
     if (words.size()!=1 || !srcfileexecution)
      cout << "askcommand will wait for a keyboard command, used in script files" << endl;
     else {
      askcommand=1;
    cout << ":"; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "givecommand")) {
     if (words.size()<2 || !srcfileexecution) {
      defineflag=1;
     cout << "givecommand will issue a given text to the parser, used in script files" << endl; }
     else
    strcpy(formula, extractfromformula(formula, 1)); }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "askvariable") || !strcmp(words[separatewords(formula, 0)].formulaWord, "input")) {
     defineflag=1;
     if (words.size()!=2) {
     cout << "askvariable/input n (or n$) will assign the next calculation to variable n/n$" << endl; }
     else {
      strcpy(askedvariable, words[separatewords(formula, 1)].formulaWord);
      askvariable=1;
      askcommand=1;
    cout << ":"; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "goto") && !srcfileexecution) {
     defineflag=1;
    cout << "goto is used in source files, to send execution to a defined label" << endl; }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "loop") && !srcfileexecution) {
     defineflag=1;
    cout << "loop is used in source files, to declare a loop beginning" << endl; }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "until") && !srcfileexecution) {
     defineflag=1;
    cout << "until is used in source files, to define a loop exit condition" << endl; }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "echo") || !strcmp(words[separatewords(formula, 0)].formulaWord, "print")) {
     defineflag=1;
     if (words.size()==1 && !srcfileexecution) {
      cout << "echo/print text..text #variable text..text $text to turn into calculation$" << endl;
      cout << "example: print value of i2:#i2 and i2*2 equals $i2*2$.useful tool!" << endl;
      cout << "print toxy(x,y) will move the cursor to coordinates" << endl;
      cout << "print color(foreground, background) changes color" << endl;
      cout << "BLACK=30 RED=31 GREEN=32 YELLOW=33 BLUE=34 MAGENTA=35 CYAN=36 WHITE=37 GREY=40" << endl;
      cout << "example: print toxy(50,10) color(34,33) Hello!" << endl;
      cout << "string functions: mid$(text,i,n), left$(text,i), right$(text,i), toupper$(text), tolower$(text)" << endl;
     cout << "example: print mid$(calculator, 1, 4) toupper$(variable$)" << endl; }
     else {
      echoparameterseparator(formula);
      cout << formula << endl;
     }
    }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "cls")) {
     defineflag=1;
     if (words.size()!=1)
      cout << "cls will clear the terminal screen" << endl;
     else
    clrscr(); }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "reset")) {
     defineflag=1;
     if (words.size()!=1)
      cout << "reset will restore the terminal ansi sequences to their originals" << endl;
     else
    resetansisequences(); }
    if (!defineflag) {
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "help") && words.size()==1) {
     defineflag=1;
     cout << "calculator recognizes regular and reverse Polish calculator syntax" << endl;
     cout << "builtin functions:abs(a), sqrt(a), sin(a), cos(a), tan(a), cotn(a), rnd(a)" << endl;
     cout << "defined functions:";
     functionslistview(-1, 1);
     cout << "interpeter commands: variable, matrix, function, precision, askcommand, givecommand, askvariable/input, evaluate, if..then..else, goto, loop..until, echo/print, cls, reset, load, run, list, alias, unalias, arbitrary, credit, help, exit" << endl;
     cout << "a single . will repeat the previous command. accepted operators: /*-+^" << endl;
     cout << "arbitrary precision:" << highlight << onoff[arbitrary] << highlight << " decimal precision output digits:" << highlight;
     if (!arbitrary)
      cout << precision;
     else
      cout << precision_a;
    cout << highlight << endl; }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "credit") && words.size()==1) {
     defineflag=1;
     cout << "calculator interpeter was written by Giorgos Saridakis." << endl <<
 "Initial code for the reverse Polish calculator routine was taken from Prentice Hall - The C Programming Language- Brian W. Kernighan, Dennis M. Ritchie, 2nd edition, chapter 4.3. Arbitrary precision arithmetics with library gmp-6.1.2." << endl;
    }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "list")) {
     defineflag=1;
     if (words.size()!=1 || !srcfilelines)
      cout << "list will list the contents of a loaded file" << endl; 
     else
     listsourcefile();
    }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "arbitrary")) {
     defineflag=1;
     if (words.size()!=1)
      cout << "arbitrary will switch arbitrary precision calculation on/off" << endl;
     else {
      arbitrary=(!arbitrary) ? 1 : 0;
    cout << "arbitrary precision:" << highlight << onoff[arbitrary] << highlight << endl; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "alias")) {
     defineflag=1;
     if (words.size()<3 && words.size()!=1) {
      cout << "alias/alias <alias name> <text> will list aliases/create an alias replacer text" << endl;
     cout << "example: alias vd variable define" << endl; }
     if (words.size()==1)
      aliaseslistview();
     if (words.size()>2) {
      n=0;
      for (i=strlen(words[separatewords(formula, 0)].formulaWord)+strlen(words[separatewords(formula, 1)].formulaWord)+2;i<strlen(formula);i++)
       tformula[n++]=formula[i];
      tformula[n]='\0';
      i=aliasesadderremover(words[separatewords(formula, 1)].formulaWord, tformula);
     if (!srcfileexecution) {
      if (!i) 
       cout << words[separatewords(formula, 1)].formulaWord << " already exists or contains numbers or is part of other function/varible" << endl;
      else
    cout << words[separatewords(formula, 1)].formulaWord << " added" << endl; } } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "unalias")) {
     defineflag=1;
     if (words.size()!=2)
      cout << "unalias <alias name> will remove an alias replacer" << endl;
     else {
      if (!(i=aliasesadderremover(words[separatewords(formula, 1)].formulaWord, tformula, 1)))
       cout << "nonexistant alias" << endl;
      else
    cout << words[separatewords(formula, 1)].formulaWord << " removed" << endl; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "precision")) {
     defineflag=1;
     if (words.size()!=2 || atoi(words[separatewords(formula, 1)].formulaWord)<0) {
      cout << "precision <a> sets decimal precision output format to a" << endl;
     cout << "in arbitrary precision mode, maximum digits will be accommodated in memory size defined in floatprecisionbits variable" << endl; }
     else {
      n=atoi((words[separatewords(formula, 1)].formulaWord));
      if (!arbitrary) {
       precision=n;
      cout.precision(precision); }
      else
       precision_a=n;
    cout << "decimal precision output format to " << highlight << n << highlight << " digits" << endl; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "load")) {
     defineflag=1;
     if (words.size()!=2) 
      cout << "load <filename> will load a source file into memory" << endl;
     else {
      strcpy(srcfilename, words[separatewords(formula, 1)].formulaWord);
      if ((i=newsourcefile(srcfilename)))
       cout << srcfilename << " loaded" << endl; 
      else
    cout << "file was not found" << endl; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "run")) {
     defineflag=1;
     if (words.size()!=1 && words.size()!=2) {
      cout << "run <filename> will load and execute a given file" << endl;
     cout << "run will execute the previously loaded source file" << endl; }
     else {
      if (words.size()==2)
       strcpy(srcfilename, words[separatewords(formula, 1)].formulaWord);
      if ((i=newsourcefile(srcfilename, words.size()-2)))
       srcfileexecution=1;
      else
       cout << "file was not found or no file in memory" << endl;
    memset(formula, 0, sizeof(formula)); } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "variable") && !strcmp(words[separatewords(formula, 1)].formulaWord, "list") && words.size()==2 && !defineflag) {
     defineflag=1;
    variableslistview(); }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "function") && !strcmp(words[separatewords(formula, 1)].formulaWord, "list") && words.size()==2 && !defineflag) {
     defineflag=1;
    functionslistview(); }
    if ((!strcmp(words[separatewords(formula, 0)].formulaWord, "variable") || !strcmp(words[separatewords(formula, 0)].formulaWord, "matrix")) && !strcmp(words[separatewords(formula, 1)].formulaWord, "remove") && !defineflag) {
     defineflag=1;
     i=removevariable(words[separatewords(formula, 2)].formulaWord);
     if (!srcfileexecution) {
      if (i>-1)
       cout << words[separatewords(formula, 2)].formulaWord << " removed" << endl; 
      else
    cout << "nonexistant variable/matrix" << endl; } }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "function") && !strcmp(words[separatewords(formula, 1)].formulaWord, "remove") && !defineflag) {
     defineflag=1;
     i=removefunction(words[separatewords(formula, 2)].formulaWord);
     if (!srcfileexecution) {
      i=(i>-1) ? 0 : i;
      switch (i) {
       case -2:
        cout << words[separatewords(formula, 2)].formulaWord << " is internal function, will not remove" << endl;
       break;
       case -1:
        cout << "nonexistant variable" << endl;
       break;
       case 0:
       cout << words[separatewords(formula, 2)].formulaWord << " removed" << endl; 
    break; } } }
    if ((!strcmp(words[separatewords(formula, 0)].formulaWord, "variable") || !strcmp(words[separatewords(formula, 0)].formulaWord, "matrix")) && words.size()<4 && !defineflag) {
     defineflag=1;
     cout << "variable define <variable name> <variable value>" << endl;
     cout << "variable define <variable name$> <variable text>" << endl;
     cout << "examples: variable define i1 5/variable define n$ Text in calculator" << endl;
     cout << "variable remove <variable name>" << endl;
     cout << "variable=variable{operator}variable" << endl;
     cout << "variable$=variable$+extra text" << endl;
     cout << "variable1$=variable1$+#variable2 sets variable with the contents of #variable2" << endl;
     cout << "variable={new value}" << endl;
     cout << "a matrix is a variable containing matrix values: matrix define network 2,5,9" << endl;
     cout << "network[1,4,2]=123.45. matrixes can be used as any other variable" << endl;
    cout << "variable list" << endl; }
    if (!strcmp(words[separatewords(formula, 0)].formulaWord, "function") && words.size()<4 && !defineflag) {
     defineflag=1;
     cout << "function define <function name> <function prototype>" << endl;
     cout << "example: function define sumcubes (#a^3)+(#b^3)" << endl;
     cout << "example: function define addfive #a+#b+#c+#d+#e" << endl;
     cout << "functions are called in typical fashion, sumcubes(a,b)" << endl;
     cout << "function remove <function name>" << endl;
    cout << "function list" << endl; }
    if ((!strcmp(words[separatewords(formula, 0)].formulaWord, "variable")  && !strcmp(words[separatewords(formula, 1)].formulaWord, "define") && !defineflag)) {
     defineflag=1;
     strcpy(tformula, words[separatewords(formula, 3)].formulaWord);
     d=runcalculationseries(tformula);
     // variable is string type ?
     if (!d || words[separatewords(formula, 2)].formulaWord[strlen(words[separatewords(formula, 2)].formulaWord-1)]=='$') {
      strcpy(tformula, words[separatewords(formula, 3, 0)].formulaWord);
      for (i=4;i<words.size();i++) {
       strcat(tformula, " ");
     strcat(tformula, words[separatewords(formula, i, 0)].formulaWord); } }
     i=definevariable(words[separatewords(formula, 2)].formulaWord, tformula, d);
     if (!srcfileexecution) {
      switch (i) {
       case 0:
        cout << words[separatewords(formula, 2)].formulaWord << " already set or part of other variable/function or used in commands" << endl;
       break;
       case -1:
        cout << words[separatewords(formula, 2)].formulaWord << " is unacceptable" << endl;
       break;
       case 1:
        cout << words[separatewords(formula, 2)].formulaWord << " set as ";
        if (words[separatewords(formula, 2)].formulaWord[strlen(words[separatewords(formula, 2)].formulaWord)-1]!='$')
         cout << d << endl;
        else
         cout << "\"" << tformula << "\"" << endl;
        break;
      }
     }
    }
    if ((!strcmp(words[separatewords(formula, 0)].formulaWord, "matrix")  && !strcmp(words[separatewords(formula, 1)].formulaWord, "define") && !defineflag)) {
     defineflag=1;
     strcpy(tformula, extractfromformula(formula, 3));
     i=definesetrecallmatrixvariable(words[separatewords(formula, 2)].formulaWord, tformula);
     switch (i) {
       case 0:
        cout << words[separatewords(formula, 2)].formulaWord << " already set or part of other variable/function or used in commands" << endl;
       break;
       case -1:
        cout << words[separatewords(formula, 2)].formulaWord << " is unacceptable" << endl;
       break;
       case -2:
        cout << words[separatewords(formula, 2)].formulaWord << " has wrongly defined (or excessive) parameter(s)" << endl;
        cout << "current limitations are dimensions:" << highlight << MAXMATRIX << highlight << " max points per matrix:" << highlight << MAXMATRIXPOINTS << highlight << endl;
       break;
       case 1:
        cout << words[separatewords(formula, 2)].formulaWord << " is a new matrix" << endl;
       break;
     }
    }
    if ((!strcmp(words[separatewords(formula, 0)].formulaWord, "function")  && !strcmp(words[separatewords(formula, 1)].formulaWord, "define") && !defineflag)) {
     defineflag=1;
     i=definefunction(words[separatewords(formula, 2)].formulaWord, words[separatewords(formula, 3)].formulaWord);
     if (!srcfileexecution) {
      switch (i) {
       case 0:
        cout << words[separatewords(formula, 2)].formulaWord << " already set or part of other variable/function or used in commands" << endl;
       break;
       case -1:
        cout << words[separatewords(formula, 2)].formulaWord << " is unacceptable" << endl;
       break;
       case -2:
        cout << words[separatewords(formula, 3)].formulaWord << " has no parameters" << endl;
       break;
       case 1:
        cout << "new function:";
        functionslistview(functions.size()-1); 
       break;
      }
     }
    }
    }
   }
   
   // entire formula to lower chars
   for (i=0;i<strlen(formula);i++)
    formula[i]=tolower(formula[i]);

   // search for variables
   negatesymbolsfromstring(formula, 1, '#');
   searchandsubstitutevariables(formula);

   // parse formula
   strcpy(tformula, formula);
   // remove symbols that might confuse reversepolishcalculator()
   negatesymbolsfromstring(tformula);
   // parse for functions
   i=parseformulaforfunctions(tformula);
   // parse for nondefined functions in syntax
   if ((formulaerrors=parseformulaforerrors(tformula, 1)) && !i && !defineflag) {
    defineflag=1;
    if (formulaerrors==-1)
   cout << (char) 34 << formula << (char) 34 << " contains nondefined function(s)" << endl; }
   // function parser result
   if (!i)
    i=parenthesesincluderforpolishreversecalculator(tformula);
   else if (!defineflag) {
    defineflag=1;
    if (i<0)
     cout << (char) 34 << formula << (char) 34 << " is not syntactically correct" << endl;
    else
   cout << (char) 34 << formula << (char) 34 << " has missing parentheses" << endl; }
   // parentheses includer result
   if (!i)
    n=reversepolishcalculatorequalizer(tformula);
   else if (!defineflag) {
    cout << formula << " has missing parentheses" << endl;
   defineflag=1; }
   // otherwise, askedvariable string is not calculated in routine below
   if (isstringvariable(askedvariable))
    n=1;
   // reversepolishcalculatorequalizer result
   if (n && !defineflag) {
    if (!arbitrary) {
    d=reversepolishcalculator(tformula); // keep for previous calculation recall
    if (!askvariable) 
     cout << "\t" << d << endl; }
    else
     d=arbitraryprecisionreversepolishcalculator(tformula);
    if (askvariable) {
     askvariable=0;
     if (isstringvariable(askedvariable))
      strcpy(emptystring, ttformula);
     removevariable(askedvariable);
     definevariable(askedvariable, emptystring, d);
     memset(askedvariable, 0, sizeof(askedvariable));
    }
   }
   else if (!defineflag) // nothing is understood
    cout << (char) 34 << formula << (char) 34 << " is undefined and/or line is semantically incorrect" << endl;
  
   }
  
 return 0;
}


