// terminal calculator

const int MAXNAME=80;
const int MAXMATRIX=5;
const int MAXMATRIXPOINTS=500;
const int MAXLOOPLINES=10000;
const int MAXSRCLINES=25000;
const int MAXTEXTREPLACEMENTS=1000; // precaution for stack smashing in inserttextpart
const int QUOTE=34;
const char *negatedsymbols="~`!@&|'\"\\[]{}';:?_";

void initiateinternalfunctions();
int definevariable(char name[MAXNAME+1], char stringcontent[], double value=0);
int removevariable(char name[MAXNAME+1]);
double recallvariable(char name[MAXNAME+1]);
void variableslistview();
void searchandsubstitutevariables(char formula[], int evaluationflag=0);
int isvariablevalid(char *name);
void searchandsubstituteprefixedvariables(char formula[]);
int separatewords(char formula[], int wordidtoreturn=0, int lowercaseflag=1);
int parseformulaforfunctions(char formula[], int stringflag=0);
int mathfunctionsparser(int function_id, char tcommand[MAXSTRING]);
void functionslistview(int function_id=-1, int flag=0);
int definefunction(char function[MAXNAME+1], char prototype[MAXCOMMAND]);
int removefunction(char name[MAXNAME+1]);
void extracttextpart(char source[MAXCOMMAND], char dest[MAXCOMMAND], int startpt, int endpt);
int inserttextpart(char text[MAXCOMMAND], char part[MAXCOMMAND], int point);
int tryfile(char *file);
int newsourcefile(char *filename, int flag=0);
int executesourcefile(char formula[]);
int loadsourcefile(char *filename);
char *getline2(ifstream& in);
int listsourcefile();
void negatesymbolsfromstring(char formula[], int char_args=0, ...);
int parseformulaforerrors(char formula[], int flag=0);
void removedoublespaces(char text[]);
int isspace(char t);
int iscomparisonsign(char t);
bool evaluatecondition(char condition1[], char condition2[], int comparisonsign=0);
int readifconditionandresult(char formula[]);
int parseformulaforvariablerecalculation(char formula[]);
void stringquotesencloser(char *tstring, int flag=0);
void stringquotesopener(char *tstring);
int isinquotes(char tstring[]);
double bringfractionpartofdivision(int param1, int param2, int scale=10);
int echoparameterseparator(char formula[]);
double runcalculationseries(char formula[]);
void limitspaces(char *tstring);
char *itoa(long int val, int base=10);
char *generaterandomvariablename(int stringflag, int namelength=8);
int isoperator(char t);
void aliasesreplacer(char formula[]);
int aliasesadderremover(char name[], char text[], int flag=0);
void aliaseslistview();
int isalreadydefined(char *name);
int stringsystemfunctionparser(int reference, char tcommand[]);
int isstringvariable(char *tvariable);
int isremline(const char formula[]);
char* extractfromformula(char formula[], int word_id);
int countparentheses(char *tstring);
int ismatrixvariable(char variable[]);
int definesetrecallmatrixvariable(char name[], char sizeparameters[], int flag=0);
int readmatrixvariableparameters(char formula[]);

// highlight text manipulator
ostream &highlight(ostream &stream)
{
  if (!highlightedtext) {
   cout << "\e[47m\e[30m";
  highlightedtext=1; }
  else {
   cout << "\e[0m" ;
  highlightedtext=0; }
  
return stream;
}

struct MatrixValue {
 double matrixValue;
vector<int> pt; } ;

// classes for variables, words and functions
class Variable {
 public:
  char variableName[MAXNAME+1];
  double variableValue;
  char variableStringcontent[MAXCOMMAND];
  vector<MatrixValue> mValue;
  vector<int> mSize;
  Variable(char s1[MAXNAME+1], char s2[MAXCOMMAND], double d) { strcpy(variableName, s1); variableValue=d; strcpy(variableStringcontent, s2); };
  Variable() { };
~Variable() { } ; } ;

struct Word { char formulaWord[MAXSTRING]; };

class Function {
 public:
  char functionName[MAXSTRING];
  char functionPrototype[MAXCOMMAND];
  int functionParameters;
  int internalflag; // 0 none, 1 arithmetic, 2 string, 3 system function
  Function(char s1[MAXSTRING], char s2[MAXCOMMAND], int i, int i1=0) { strcpy(functionName, s1); strcpy(functionPrototype, s2); functionParameters=i; internalflag=i1; };
  Function(const char s1[MAXSTRING], const char s2[MAXCOMMAND], int i, int i1=0) { strcpy(functionName, s1); strcpy(functionPrototype, s2); functionParameters=i; internalflag=i1; };
  Function() { } ;
~Function() { }; };

class Looppoint {
 public:
  int loopLine;
  char loopLabel[MAXNAME+1]; 
  Looppoint(int i, char s[]) { loopLine=i; strcpy(loopLabel, s); }; 
  Looppoint() { };
~Looppoint() { }; };

class Alias {
 public:
  char aliasPrototype[MAXNAME+1];
  char aliasText[MAXCOMMAND]; 
  Alias(char s1[MAXNAME+1], char s2[MAXCOMMAND]) { strcpy(aliasPrototype, s1); strcpy(aliasText, s2); };
  Alias() { };
~Alias() { }; };

vector<Variable> variables;
vector<Word> words;
vector<Function> functions;
vector<Looppoint> gotolabels;
vector<Looppoint> looppoints;
vector<string> sourcefile;
vector<Alias> aliases;

// global library variables
const char *booleans[]={ "false", "true" };
const char *comparisonsigns[]={ "=", "<", ">", "<=", ">=", "<>" };
const char *nonallowednametypes[]={ "if", "then", "else", "echo", "print", "cls", "resetscreen", "variable", "matrix", "function", "precision", "askcommand", "givecommand", "askvariable", "input", "evaluate", "if", "then", "else", "echo", "loop", "until", "print", "load", "run", "list", "alias", "unalias", "arbitrary", "credit", "help", "exit" };
const char operators[]={ '+', '-', '*', '/', '^' };
const char *onoff[]={ "off", "on" };
char emptystring[MAXNAME+1];
char srcfilename[MAXNAME+1];
int srcfilelineexecuted=0, srcfilelines, insidelooplines=0;
int textreplacements=0;
hash<string> stringvariablehashvalue;
char functionParameter[MAXFUNCTIONPARAMETERS][MAXCOMMAND];

// initiate functions
void initiateinternalfunctions()
{
  // arithmetic functions
  Function tfunction1("abs", "(#a 1 !)", 1, 1);
  functions.push_back(tfunction1);
  Function tfunction2("sin", "(#a 1 [)", 1, 1);
  functions.push_back(tfunction2);
  Function tfunction3("cos", "(#a 1 ])", 1, 1);
  functions.push_back(tfunction3);
  Function tfunction4("tan", "(#a 1 ;)", 1, 1);
  functions.push_back(tfunction4);
  Function tfunction5("cotn", "(#a 1 |)", 1, 1);
  functions.push_back(tfunction5);
  Function tfunction6("sqrt", "(#a 1 ?)", 1, 1);
  functions.push_back(tfunction6);
  Function tfunction7("log", "(#a 1 :)", 1, 1);
  functions.push_back(tfunction7);
  Function tfunction8("rnd", "(#a 1 &)", 1, 1);
  functions.push_back(tfunction8);
  
  // string functions
  Function sfunction1("mid$", "(#a #b)", 2, 2);
  functions.push_back(sfunction1);
  Function sfunction2("left$", "(#a)", 1, 2);
  functions.push_back(sfunction2);
  Function sfunction3("right$", "(#a)", 1, 2);
  functions.push_back(sfunction3);
  Function sfunction4("toupper$", "(#a)", 0, 2);
  functions.push_back(sfunction4);
  Function sfunction5("tolower$", "(#a)", 0, 2);
  functions.push_back(sfunction5);
  
  // system functions
  Function yfunction1("toxy", "(#a #b)", 2, 3);
  functions.push_back(yfunction1);
  Function yfunction2("color", "(#a #b)", 2, 3);
  functions.push_back(yfunction2);
}

// define variable
int definevariable(char name[], char stringcontent[], double value)
{
  int i;
  
  for (i=0;i<strlen(name);i++)
   name[i]=tolower(name[i]);
  
   // check for identical variable or function
   if (isalreadydefined(name))
    return 0;

   // check for unacceptable characters
   for (i=0;i<strlen(name);i++)
    if (!isalpha(name[i]) && !isdigit(name[i]) && !isstringvariable(name)) // allow string variables
     return -1;
   // is variable of string type ? generate hash value
   if (isstringvariable(name)) {
    searchandsubstituteprefixedvariables(stringcontent);
    parseformulaforfunctions(stringcontent, 1);
   value=(double) stringvariablehashvalue(stringcontent); }
    
   Variable tvariable(name, stringcontent, value);
   variables.push_back(tvariable);
  
 return 1;
}

// remove variable
int removevariable(char name[MAXNAME+1])
{
  int i;
  vector<Variable>::iterator p=variables.begin();
  
   for (i=0;i<variables.size();i++)
    if (!strcmp(name, variables[i].variableName))
     break;
   if (i==variables.size())
    return -1;
    
   p+=i;
   variables.erase(p);
   
  return i;
}

// recall variable
double recallvariable(char tvariable[])
{
  int i;
  
  for (i=0;i<variables.size();i++)
   if (!strcmp(tvariable, variables[i].variableName))
    break;
  if (i==variables.size())
   return 0;

 return variables[i].variableValue;
}

// listview of variables
void variableslistview()
{
  int i, n;
  
   if (!variables.size())
    cout << "variable stack is empty" << endl;
   for (i=0;i<variables.size();i++) {
       
    if (!isstringvariable(variables[i].variableName)) {
    cout << variables[i].variableName << " ";
    if (!ismatrixvariable(variables[i].variableName))
     cout << variables[i].variableValue << endl;
    else {
     cout << "[";
     for (n=0;n<variables[i].mSize.size();n++) {
      cout << variables[i].mSize[n];
      if (n<variables[i].mSize.size()-1)
       cout << ",";
      else
   cout << "]" << endl; } } }
   else  
    cout << variables[i].variableName << " \"" << variables[i].variableStringcontent << "\"" << endl;
   
   }
}

// extract and replace variable in string
void searchandsubstitutevariables(char formula[], int evaluationflag)
{
  int i, operation=1, startpt, endpt;
  char tvariable[MAXCOMMAND];
  evaluationflag=(evaluationflag>1) ? 0 : evaluationflag;

   textreplacements=0;
   while (operation) {
       
    operation=0;
    for (i=0;i<variables.size();i++) {
     if ((startpt=findsimple(formula, variables[i].variableName))) {
      operation=1;
      endpt=startpt+strlen(variables[i].variableName)-1;
      if (ismatrixvariable(variables[i].variableName)) {
       while (formula[endpt]!=']' && endpt<strlen(formula))
        ++endpt;
       ++endpt;
       extracttextpart(formula, tvariable, startpt-1, endpt);
       readmatrixvariableparameters(tvariable);
       strcpy(tvariable, dtoa(variables[i].variableValue));
      inserttextpart(formula, tvariable, startpt-1); }
      else { 
      extracttextpart(formula, tvariable, startpt-1, endpt);
      if (isstringvariable(variables[i].variableName) && !evaluationflag)
       operation=inserttextpart(formula, variables[i].variableStringcontent, startpt-1);
      if (isstringvariable(variables[i].variableName) && evaluationflag==1)
       operation=inserttextpart(formula, dtoa(recallvariable(tvariable)), startpt-1);
      if (!isstringvariable(variables[i].variableName))
       operation=inserttextpart(formula, dtoa(recallvariable(tvariable)), startpt-1); 
      }
     }
    }
    
   }
}

// is variable found
int isvariablevalid(char *name)
{
  int i;

   for (i=0;i<(int)variables.size();i++)
    if ( !strcmp(variables[i].variableName, name) )
     break;

   if ( i == (int)variables.size() )
    return 0;

 return 1;
}

// extract and replace variable in string
void searchandsubstituteprefixedvariables(char formula[])
{
  int i, operation=1, pt[2];
  char tformula[MAXCOMMAND], textract[MAXCOMMAND], tvariable[MAXNAME+1];
  strcpy(tformula, formula);  
  
   // parse formula, replace variables that have # before
   textreplacements=0;
   while (operation) {
       
    operation=0;
    for (i=0;i<variables.size();i++) {
     strcpy(tvariable, "#");
     strcat(tvariable, variables[i].variableName);
     if ((pt[0]=findsimple(tformula, tvariable))) {
      pt[1]=pt[0]+strlen(variables[i].variableName);
      extracttextpart(tformula, textract, pt[0]-1, pt[1]);
      if (!isstringvariable(variables[i].variableName))
       operation=inserttextpart(tformula, dtoa(recallvariable(variables[i].variableName)), pt[0]-1);
      else
       operation=inserttextpart(tformula, variables[i].variableStringcontent, pt[0]-1); 
     }
    }
    
   }
    
  strcpy(formula, tformula);
}

// separate words in string and store in vector inside structures
int separatewords(char formula[], int wordidtoreturn, int lowercaseflag)
{
  int i, n;
  char tformula[MAXSTRING];
  Word tword;
  vector<Word>::iterator p=words.begin();
  while (p<words.end()) // because words.clear() only moves pointer, words are kept
   words.erase(p++);
  words.clear();
    
   for (i=0;i<strlen(formula);i++) {
       
    n=0;
    while (!isspace(formula[i]) && i<strlen(formula))
     tformula[n++]=(!lowercaseflag) ? formula[i++] : tolower(formula[i++]);
    tformula[n]='\0';
    strcpy(tword.formulaWord, tformula);
    words.push_back(tword);
    
   }

  return wordidtoreturn;
}

// parse formula for user functions and call identify
int parseformulaforfunctions(char formula[], int stringflag)
{
  int i, n, startpt, endpt, findinstructions=1, parseresult;
  char ttext[MAXSTRING], tcommand[MAXCOMMAND];
  strcpy(ttext, formula);
  
   while (findinstructions) {
       
    findinstructions=0;
    for (i=0;i<functions.size();i++)
     if ((startpt=findsimple(ttext, functions[i].functionName)))
      break;
     if (i<functions.size()) {
      findinstructions=1;
      --startpt;
      n=startpt;
      // empty function callname so it will not be parsed again
      while (ttext[n]!='(' && n<strlen(ttext))
       ttext[n++]=SPACE;
      endpt=n+1;
      openparentheses=1;
      while (openparentheses && endpt<strlen(ttext)) {
       if (ttext[endpt]=='(')
        ++openparentheses;
       if (ttext[endpt]==')')
        --openparentheses;
      ++endpt; }
      if (openparentheses) // not all parentheses closed
       return openparentheses;
      extracttextpart(ttext, tcommand, startpt, endpt);
      // call routine recursively
      n=parseformulaforfunctions(tcommand);
      if (n)
       return n;
      // parse tcommand
      if (functions[i].internalflag<2 && !stringflag)
       parseresult=mathfunctionsparser(i, tcommand);
      else if (stringflag)
       parseresult=stringsystemfunctionparser(i, tcommand);
      if (parseresult)
       return parseresult;
      // now insert tcommand into startpt
      inserttextpart(ttext, tcommand, startpt);
     }
     
   }
   
  strcpy(formula, ttext);
      
 return 0;
}

// extract user function parameters and replace in functionPrototype
int mathfunctionsparser(int function_id, char tcommand[MAXSTRING])
{
  int i, mpos=0, tp=0, n, operation=1;
  double d;
  char tparameter[MAXFUNCTIONPARAMETERS][MAXSTRING];
  
   while (tcommand[mpos]!='(')
    ++mpos;
   ++mpos;
   tcommand[strlen(tcommand)-1]=',';
   for (i=0;i<functions[function_id].functionParameters;i++) {
       
    n=0;
    while (mpos<strlen(tcommand)) {
     if (tcommand[mpos]==',') {
      ++mpos;
      break;
     }
     if (tcommand[mpos]=='(')  // these were 0 when routine was called
      ++openparentheses;
     if (tcommand[mpos]==')')
      --openparentheses;
     tparameter[tp][n++]=tcommand[mpos];
     ++mpos;
    }
    tparameter[tp][n]='\0';
    if (!strlen(tparameter[tp])) { // nonexistant parameter
     tcommand[strlen(tcommand)-1]=')';
    return tp*-1; } // negative return to separate from openparentheses
    parenthesesincluderforpolishreversecalculator(tparameter[tp]);
    reversepolishcalculatorequalizer(tparameter[tp]);
    d=reversepolishcalculator(tparameter[tp]);
    strcpy(tparameter[tp], dtoa(d));
    ++tp; }
    tcommand[strlen(tcommand)-1]=')';
    if (mpos<strlen(tcommand)-1) // still more tcommand when functions have been read
     return tp*-1;
    // reconstruct tcommand
    strcpy(tcommand, functions[function_id].functionPrototype);
    searchandsubstitutevariables(tcommand);
    while (operation) {
     operation=0;
     for (i=0;i<functions[function_id].functionParameters;i++) {
      for (n=0;n<strlen(tcommand);n++)
       if (tcommand[n]==i+97 && tcommand[n-1]=='#') // find a,b,c,d,e ..MAXFUNCTIONPARAMETERS
        break;
       if (n<strlen(tcommand)) {
        operation=1;
        tcommand[n-1]=SPACE;
        tcommand[n]=SPACE;
        inserttextpart(tcommand, tparameter[i], n-1);
       }
     }
     
    }
    
 return 0;
}

// listview of functions
void functionslistview(int function_id, int flag) // 1 show only functionName
{
  int i, n, internalfunctions=0, start, end=functions.size();
  char c;
  
   for (i=0;i<functions.size();i++)
    if (functions[i].internalflag)
     ++internalfunctions;
   if (function_id>-1) {
    start=function_id; 
   end=function_id+1; }
   else
    start=internalfunctions;
   if (!(functions.size()-internalfunctions)) {
    cout << "no user defined functions";
   if (!flag)
    cout << endl;
   }
   for (i=start;i<end;i++) {
    cout << functions[i].functionName << "(" ;
    for (n=0;n<functions[i].functionParameters;n++) {
     cout << (c=n+97);
     if (n<functions[i].functionParameters-1)
    cout << ","; }
    cout << ")";
    if (!flag)
     cout << " ";
    if (flag && i<end-1)
     cout << ", ";
    if (!flag) 
     cout << functions[i].functionPrototype << endl;
   }
   if (flag)
    cout << endl;
}

// define user function
int definefunction(char function[MAXNAME+1], char prototype[MAXCOMMAND])
{
  int i, n, existingparameters[MAXFUNCTIONPARAMETERS], parameters=0;
  double d;
  
   // check if identical variable or function name exists
   if (isalreadydefined(function))
    return 0;
   // see if non alphanumerical characters exist in name
   for (i=0;i<strlen(function);i++)
    if (!isalpha(function[i]) && !isdigit(function[i]))
     break;
   if (i<strlen(function))
    return -1;
   // zero to all array positions
   for (i=0;i<MAXFUNCTIONPARAMETERS;i++)
    existingparameters[i]=0;
   // check prototype for a,b,c,d,e...
   for (n=0;n<MAXFUNCTIONPARAMETERS;n++)
    for (i=0;i<strlen(prototype);i++)
     if (prototype[i]==n+97 && prototype[i-1]=='#')
      existingparameters[n]=1;
   // number of function parameters
   for (i=0;i<MAXFUNCTIONPARAMETERS;i++)
    if (existingparameters[i])
     ++parameters;
    else
     break;
   if (!parameters)
    return -2;
   // add to functions
   Function tfunction(function, prototype, parameters);
   functions.push_back(tfunction);
    
 return 1;
}

// remove function
int removefunction(char name[MAXNAME+1])
{
  int i;
  vector<Function>::iterator p=functions.begin();
  
   for (i=0;i<functions.size();i++)
    if (!strcmp(name, functions[i].functionName))
     break;
   if (functions[i].internalflag)
    return -2;
   if (i==functions.size())
    return -1;
    
   p+=i;
   functions.erase(p);
   
  return i;
}

// remove part of text and place in a char array
void extracttextpart(char source[], char dest[], int startpt, int endpt)
{
  int i, n=0;
  char ttext[MAXCOMMAND];
  
   for (i=startpt;i<endpt;i++)
    dest[n++]=source[i];
   dest[n]='\0';
   n=0;
   for (i=0;i<strlen(source);i++)
    if (i<startpt || i>=endpt)
     ttext[n++]=source[i];
   ttext[n]='\0';
   strcpy(source, ttext);
}

// insert text in text
int inserttextpart(char text[], char part[], int point)
{
  char ttext[MAXCOMMAND];
  int i, n=0;
  
   // precaution so final string will not exceed MAXCOMMAND
   if ((strlen(text)+strlen(part))>=MAXCOMMAND || textreplacements>MAXTEXTREPLACEMENTS)
    return 0;
   for (i=0;i<point;i++)
    ttext[n++]=text[i];
   for (i=0;i<strlen(part);i++)
    ttext[n++]=part[i];
   for (i=point;i<strlen(text);i++)
    ttext[n++]=text[i];
   ttext[n]='\0';
   strcpy(text, ttext);
 
 return 1;
}

// int try if file exists
int tryfile(char *file) 
{
  FILE *f;
   
  f=fopen(file, "r");
  if (!f)
   return 0;
  fclose (f);
  
 return 1;
}

// new sourcefile
int newsourcefile(char *filename, int flag) // 0 load new file, 1 keep old
{
   // clear variables
   srcfilelineexecuted=0; 
   // flag=1, keep previously loaded lines
   if (flag==1)
    return 1;
   srcfilelines=0;
   // see if file is there and clear vectors
   if (!tryfile(filename))
    return 0;
   sourcefile.clear();
   looppoints.clear();
   gotolabels.clear();
   // call loadsourcefile
   loadsourcefile(filename);
    
 return 1;
}

// load sourcefile
int loadsourcefile(char *filename)
{
  int i;
  char line[MAXCOMMAND], tname[MAXNAME+1];
  ifstream insrcfile(filename);
   
   // load file into string vector
   while ( insrcfile ) {
    insrcfile.getline(line, MAXCOMMAND);
//     strcpy(line, getline2(insrcfile));
    limitspaces(line);
    string tstring=line;
    sourcefile.push_back(tstring); 
     ++srcfilelines;
   }
   
   // parse for line labels
   for (i=0;i<srcfilelines;i++) {
       
    strcpy(line, sourcefile[i].c_str());
    if (words[separatewords(line)].formulaWord[strlen(words[separatewords(line)].formulaWord)-1]==':' && words.size()==1) {
     strcpy(tname, words[separatewords(line)].formulaWord);
     tname[strlen(tname)-1]='\0';
     Looppoint tlabel(i, tname);
     gotolabels.push_back(tlabel);
    }
    
   }

  insrcfile.close();
    
 return 1;
}

// simplified getline
char *getline2(ifstream& in)
{
  static char tline[MAXCOMMAND];
  int c, i;
  c=i=0;
  
   while ( in && c != '\n' )
    tline[i++]=c=in.get();
   tline[i]='\0';
    
 return &tline[0];   
}

// classic list source file
int listsourcefile()
{
  int i;
  if (!srcfilelines)
   return 0;
  
   for (i=0;i<srcfilelines-1;i++)
    cout << sourcefile[i] << endl;
    
 return 1;
}

// read .src file and send to command line
int executesourcefile(char formula[])
{
  int i, n;
  char c, tname[MAXNAME+1], tformula[MAXCOMMAND];
  vector<Looppoint>::iterator p;
  
  // if end of file has been reaches, reset variables
  if ( srcfilelineexecuted == srcfilelines - 1 )
   return 0;
  
  // read current line
  strcpy(formula, sourcefile[srcfilelineexecuted].c_str());
  
  // does formula contain if..then ?
  readifconditionandresult(formula);
  
  // parse for goto statement
  if (!strcmp(words[separatewords(formula)].formulaWord, "goto") && words.size()==2) {
   defineflag=1;
   strcpy(tname, words[separatewords(formula, 1)].formulaWord);   
   for (i=0;i<gotolabels.size();i++) {
    if (!strcmp(gotolabels[i].loopLabel, tname)) {
     srcfilelineexecuted=gotolabels[i].loopLine+1; {
      return 1;
     }
    }
   }
  }
  
  // parse for labels and skip
  if (words[separatewords(formula)].formulaWord[strlen(words[separatewords(formula)].formulaWord)-1]==':' && words.size()==1)
   defineflag=1;
   
  // parse for loop starts
  if (!strcmp(words[separatewords(formula)].formulaWord, "loop") && words.size()==1) {
   defineflag=1;
   strcpy(tname, "loop");
   strcat(tname, itoa(looppoints.size()+1));
   Looppoint tloop(srcfilelineexecuted, tname);
   looppoints.push_back(tloop);
  }
  // parse for until... statements
  if (!strcmp(words[separatewords(formula)].formulaWord, "until") && looppoints.size()) {
      
   defineflag=1;
   n=0;
   for (i=5;i<strlen(formula);i++)
    tformula[n++]=formula[i];
   tformula[n]='\0';
   strcpy(formula, "if ");
   strcat(formula, tformula);
   strcat(formula, " then 1 else 0");
   if ((n=readifconditionandresult(formula))) {
    if (!strcmp(formula, "0")) {
     srcfilelineexecuted=looppoints[looppoints.size()-1].loopLine+1;
     ++insidelooplines;
     if (insidelooplines>MAXLOOPLINES) {
      cout << "it appears execution has been caught in an infinite loop.break ?";
      char c=getchar();
      if (tolower(c)=='y')
       n=-1;
      else {
       insidelooplines=0;
     return 1; } }
     if (n!=-1)
    return 1; }
    insidelooplines=0;
    p=looppoints.end();
    --p;
   looppoints.erase(p);
   }
   
  }

  // prepare to leave
  ++srcfilelineexecuted;
  
 return 1;
}

int sum(int num_args, ...) {
   int val = 0;
   va_list ap;
   int i;

   va_start(ap, num_args);
   for(i = 0; i < num_args; i++) {
      val += va_arg(ap, int);
   }
   va_end(ap);
 
   return val;
}

// negate symbols from line input
void negatesymbolsfromstring(char formula[], int char_args, ...)
{
  int i, i1=(int)strlen(negatedsymbols);
  va_list ap;

  char tnegatedsymbols[MAXNAME+1];
  strcpy( tnegatedsymbols, negatedsymbols );
  va_start(ap, char_args);
  for (i=0;i<char_args;i++)
   tnegatedsymbols[i1++]=va_arg(ap, int);
  tnegatedsymbols[i1]='\0';
      
   for (i=0;i<strlen(formula);i++)
    for (i1=0;i1<strlen(tnegatedsymbols);i1++)
     if ( formula[i] == tnegatedsymbols[i1] )
      formula[i]=SPACE;
}

// parse formula for nondefined functions and too long numbers
int parseformulaforerrors(char formula[], int flag) // 0 both, 1 functions, 2 numbers
{
  int i, n;

   if (!flag || flag==1) {
    for (i=0;i<strlen(formula);i++) {
     if (isalpha(formula[i]) && (formula[i+1]=='(' || formula[i+1]=='('))
      return -1; // nondefined function
    }
   }
   if (!flag || flag==2) {
    for (i=0;i<strlen(formula);i++) {
     if (isdigit(formula[i])) {
      n=0;
      while ((isdigit(formula[i]) || isdecimalseparator(formula[i])) && i<strlen(formula)) {
       ++n;
      ++i; }
     if (n>MAXOP-1)
      return -2; // too long number
     }
    }
   }
   
 return 0;
}

// remove unnecessary spaces from string
void removedoublespaces(char text[])
{
  int i=0, n=0;
  char ttext[strlen(text)];

   while (isspace(text[i]))
    ++i;
   for (;i<strlen(text);i++) {
    while (isspace(text[i]) && isspace(text[i+1]) && i<strlen(text))
     ++i;
    ttext[n++]=text[i];
   }
   ttext[n]='\0';
  
  strcpy(text, ttext);
}

// isspace
int isspace(char t)
{
  if ( t == 32 )
   return 1;
    
 return 0;
}

// iscomparison sign
int iscomparisonsign(char t)
{
  int i;
  char tcomparisonsigns[]={ '=', '<', '>' };
  
   for (i=0;i<3;i++)
    if (t==tcomparisonsigns[i])
     return 1;
    
  return 0;
}

// evaluate condition, entire sequences for calculator function included
bool evaluatecondition(char condition1[], char condition2[], int comparisonsign)
{
  int i, n, nomoreparse=0, formulaerrors;
  double d[2];
  char tevaluation[MAXCOMMAND];
  
   for (n=0;n<2;n++) {
    if (!n)
     strcpy(tevaluation, condition1);
    else     
     strcpy(tevaluation, condition2);
    if ((formulaerrors=parseformulaforerrors(tevaluation, 2)))
     return 0;
    removedoublespaces(tevaluation);
    parseformulaforfunctions(tevaluation, 1);
    searchandsubstitutevariables(tevaluation, comparisonsign+1); // only = will work in string hash value comparison
    negatesymbolsfromstring(tevaluation);
    // parse for functions
    i=parseformulaforfunctions(tevaluation);
    if (i)
     return 0;
    // is routine asked to evaluate text out of variables ?
    nomoreparse=0;
    if (!atof(tevaluation) && isalpha(tevaluation[0]) && !comparisonsign) {
     nomoreparse=1;
     if (n==1) // remove space from second condition
      tevaluation[strlen(tevaluation)-1]='\0';
    d[n]=(double) stringvariablehashvalue(tevaluation); }
    if (!nomoreparse) {
    // parse for nondefined functions in syntax
    if ((i=parseformulaforerrors(tevaluation, 1)))
     return 0;
    // function parser result
    i=parenthesesincluderforpolishreversecalculator(tevaluation);
    if (i)
     return 0;
    // parentheses includer result
    i=reversepolishcalculatorequalizer(tevaluation);
    if (!i)
     return 0;
    d[n]=reversepolishcalculator(tevaluation);
    }
   }
    
    // compare values
    switch (comparisonsign) {
     case 0:
      if (d[0]!=d[1])
       return 0;
     break;
     case 1:
      if (d[0]>=d[1])
       return 0;
     break;
     case 2:
      if (d[0]<=d[1])
       return 0;
     break;
     case 3:
      if (d[0]>d[1])
       return 0;
     break;
     case 4:
      if (d[0]<d[1])
       return 0;
     break;
     case 5:
      if (d[0]==d[1])
       return 0;
     break;
    }
    
  return 1; // true
}

// read if condition and result
int readifconditionandresult(char formula[])
{
  int i, n, elseflag=0, startpt[3], endpt, comparisonsign;
  char condition[2][MAXCOMMAND], result[MAXCOMMAND], elsecondition[MAXCOMMAND], comparisonrule[3], tformula[MAXCOMMAND];
  strcpy(tformula, formula);
  
    // find if and then and extract
   if ((startpt[0]=findsimple(tformula, const_cast<char *> ("if")))!=1)
    return -1;
   if (!(startpt[1]=findsimple(tformula, const_cast<char *> ("then"))))
    return -1;
   extracttextpart(tformula, condition[0], startpt[0]+2, startpt[1]-1);
   if (findsimple(condition[0], const_cast<char *> ("if")))
    return -1;
   startpt[1]=findsimple(tformula, const_cast<char *> ("then"));
   // is there else in formula ?
   startpt[2]=strlen(tformula);
   if ((elseflag=findsimple(tformula, const_cast<char *> ("else"))))
    startpt[2]=elseflag-1;
   endpt=startpt[2];
   extracttextpart(tformula, result, startpt[1]+4, endpt);
   if (findsimple(result, const_cast<char *> ("if")) || findsimple(result, const_cast<char *> ("then")))
    return -1;
   if (elseflag) {
    startpt[2]=findsimple(tformula, const_cast<char *> ("else"));
    extracttextpart(tformula, elsecondition, startpt[2]+4, strlen(tformula)); 
    if (findsimple(result, const_cast<char *> ("if")) || findsimple(result, const_cast<char *> ("then")) || findsimple(result, const_cast<char *> ("else")))
    return -1;
   }
   // find and extract comparison rules, detect wrong syntax
   for (i=0;i<strlen(condition[0]);i++)
    if (iscomparisonsign(condition[0][i]))
     break;
   if (i==strlen(condition[0]))
    return -1;
   comparisonrule[0]=condition[0][i];
   condition[0][i]=SPACE;
   if (iscomparisonsign(condition[0][i+1]) && comparisonrule[0]!='=') {
    comparisonrule[1]=condition[0][i+1];
    comparisonrule[2]='\0';
    condition[0][i+1]=SPACE;
   }
   else
    comparisonrule[1]='\0';
   // search again and return 0 if more signs are found
   for (n=0;n<strlen(condition[0]);n++)
    if (iscomparisonsign(condition[0][n]))
     return -1;
   // extract second part of condition
   extracttextpart(condition[0], condition[1], i, strlen(condition[0]));
   for (comparisonsign=0;comparisonsign<6;comparisonsign++)
    if (!strcmp(comparisonsigns[comparisonsign], comparisonrule))
     break;
   // evaluate formula
   i=evaluatecondition(condition[0], condition[1], comparisonsign);
   // run result
   if (i)
    strcpy(formula, result);
   else if (elseflag) {
    i=1;
   strcpy(formula, elsecondition); }
    
  return i;
}

// parse formula for variable recalculation, ex. i=i+1
int parseformulaforvariablerecalculation(char formula[])
{
  int i, n, l, startpt=0;
  double d;
  char tformula[2][MAXCOMMAND], ttext[MAXNAME+1];
  strcpy(tformula[0], formula);
  
   // remove starting spaces
   while (isspace(tformula[0][startpt]))
    ++startpt;
   // see if we have variable in the start of calculation formula
   for (n=0;n<variables.size();n++)
    if ((startpt=findsimple(tformula[0], variables[n].variableName))==1)
     break;
   if (n==variables.size())
    return 0;
   // if it's a matrix variable, remove parameters to ttext
   if (ismatrixvariable(variables[n].variableName)) {
    l=startpt+strlen(variables[n].variableName);
    while (tformula[0][l]!=']' && l<strlen(tformula[0]))
     ++l;
    extracttextpart(tformula[0], ttext, startpt+strlen(variables[n].variableName)-1, l+1);
    startpt=findsimple(tformula[0], variables[n].variableName)+strlen(variables[n].variableName)-1;
   }
   else
    startpt+=strlen(variables[n].variableName)-1;
   // is there a = after first variable
   if (tformula[0][startpt]!='=')
    return 0;
   // extract text after =
   extracttextpart(tformula[0], tformula[1], startpt+1, strlen(tformula[0]));
   // is variable of string type ?
   if (isstringvariable(variables[n].variableName)) {
    if ((i=findsimple(tformula[1], variables[n].variableName)))
     for (i=0;i<strlen(variables[n].variableName);i++)
      tformula[1][i]=SPACE;
    limitspaces(tformula[1]);
    searchandsubstituteprefixedvariables(tformula[1]);
    parseformulaforfunctions(tformula[1], 1);
    if (tformula[1][0]!='+')
     strcpy(variables[n].variableStringcontent, tformula[1]);
    else {
     extracttextpart(tformula[1], emptystring, 0, 1); // remove the +
    strcat(variables[n].variableStringcontent, tformula[1]); } 
    strcpy(tformula[1], variables[n].variableName);
    strcpy(emptystring, variables[n].variableStringcontent);
    removevariable(variables[n].variableName);
    definevariable(tformula[1], emptystring);
   }
   else {
    // reconstruct tformula[0]
    limitspaces(tformula[1]);
    strcpy(tformula[0], tformula[1]);
    // run calculation series
    variables[n].variableValue=runcalculationseries(tformula[0]);
    if (ismatrixvariable(variables[n].variableName)) {
     ttext[0]=SPACE; 
     ttext[strlen(ttext)-1]=SPACE;
     limitspaces(ttext);
     definesetrecallmatrixvariable(variables[n].variableName, ttext, 2);
    }
   }
   
  return 1;
}

// enclose string in quotes, add ; if wanted
void stringquotesencloser(char *tstring, int flag) // 1 add ; at the end of string
{
  int i;
  char ttstring[MAXCOMMAND];
  ttstring[0]=QUOTE;
 
   for (i=0;i<strlen(tstring);i++)
    ttstring[i+1]=tstring[i];
   ++i;
   ttstring[i++]=QUOTE;
   if (flag)
    ttstring[i++]=';';
   ttstring[i]='\0';
   strcpy(tstring, ttstring);
}

// remove quotes from string
void stringquotesopener(char *tstring)
{
  int i, n=0;
  char ttstring[MAXCOMMAND];
 
   for (i=1;i<strlen(tstring)-1;i++)
    ttstring[n++]=tstring[i];
   ttstring[n]='\0';
   strcpy(tstring, ttstring);
}

// is string in quotes ?
int isinquotes(char tstring[])
{
  if ( tstring[0] == QUOTE && tstring[strlen(tstring)-1] == QUOTE )
   return 1;
         
 return 0;
}

// divide param1/param2 and bring fraction part*scale
double bringfractionpartofdivision(int param1, int param2, int scale)
{
  double param = (double) param1/param2;
  double fractpart, intpart;
  
  fractpart = modf (param , &intpart);
 return fractpart*scale;
}

// echo segment separator 
int echoparameterseparator(char formula[])
{
  int i=0, n=0, operation=1, pt[2]={ 0, 0 };
  double d;
  char tformula[MAXCOMMAND], ttformula[MAXCOMMAND], textract[MAXCOMMAND], tvariable[MAXNAME+1];
  strcpy(tformula, formula);
  
   // remove echo or print
   if ((pt[0]=findsimple(tformula, const_cast<char *> ("echo")))!=1 &&  (pt[1]=findsimple(tformula, const_cast<char *> ("print")))!=1)
    return 0;
   i=(pt[0]==1) ? 5: 6;
   for (;i<strlen(tformula);i++)
    ttformula[n++]=tformula[i];
   ttformula[n]='\0';
   strcpy(tformula, ttformula);
   
   // parse for string functions
   parseformulaforfunctions(tformula, 1);
   
   // parse formula, replace variables that have # before
   searchandsubstituteprefixedvariables(tformula);
    
   // parse formula, replace text inside $..$ for calculation
   textreplacements=0;
   operation=1;
   while (operation) {
    operation=0;
    for (pt[0]=0;pt[0]<strlen(tformula);pt[0]++) {
     if (tformula[pt[0]]=='$') {
      pt[1]=pt[0]+1;
      while (tformula[pt[1]]!='$' && pt[1]++<strlen(tformula));
      if (pt[1]==strlen(tformula))
       return 0;
      extracttextpart(tformula, textract, pt[0], pt[1]+1);
      d=runcalculationseries(textract);
      if (d)
       operation=inserttextpart(tformula, dtoa(d), pt[0]);
     }
    }
   }
   
  strcpy(formula, tformula);
    
 return 1;   
}

// reverse Polish calculator series
double runcalculationseries(char formula[])
{
  int i;
  double d;
  char tformula[MAXCOMMAND];
  strcpy(tformula, formula);
  
    // run calculation series
   removedoublespaces(tformula);
   searchandsubstitutevariables(tformula);
   // parse for functions
   i=parseformulaforfunctions(tformula);
   if (i)
    return 0;
   // parse for nondefined functions in syntax
   if ((i=parseformulaforerrors(tformula, 1)))
    return 0;
   // function parser result
   i=parenthesesincluderforpolishreversecalculator(tformula);
   if (i)
    return 0;
   // parentheses includer result
   i=reversepolishcalculatorequalizer(tformula);
   if (!i)
    return 0;
   d=reversepolishcalculator(tformula);
    
  strcpy(formula, tformula);
   
 return d;
}

// trim spaces at start and end of string
void limitspaces(char *tstring)
{
  int i=0, n=0;
  char ttstring[MAXCOMMAND];
  
  while (isspace(tstring[i]) && i<strlen(tstring))
   ++i;
  for (;i<strlen(tstring);i++)
   ttstring[n++]=tstring[i];
  ttstring[n]='\0';
  i=strlen(ttstring)-1;
  while (isspace(ttstring[i]))
   --i;
  ttstring[++i]='\0';
  if (!strlen(ttstring))
   strcpy(ttstring, " ");
  strcpy(tstring, ttstring);
}

// integer to string
char *itoa(long int val, int base)
{
 	static char buf[32] = {0};
	int sign_flag=0, v = 30;
    buf[31]='\0';
    
	if (!val) {
     buf[v]='0';
    return &buf[v]; }
    if (val<0) {
     val*=-1;
    sign_flag=1; }
    
	for(; val && v ; --v, val /= base)
	 buf[v] = "0123456789abcdef"[val % base];
    if (sign_flag) {
     buf[v]='-';
	return &buf[v]; }
	else
     return &buf[v+1];
}

// generate random variable name
char *generaterandomvariablename(int stringflag, int namelength)
{
  int i, r, operation=1;
  static char variablename[MAXNAME+1];
  
  while (operation) {
   operation=0;
   for(i=0;i<namelength;i++) {
    while ((r=rand() % 122 + 1)<97);
   variablename[i]=(char) r; }
   if (stringflag)
    variablename[i++]='$';
   variablename[i]='\0';
   for (i=0;i<variables.size();i++)
    if (!strcmp(variablename, variables[i].variableName))
  operation=1; }
    
 return &variablename[0];
}

// is character operator
int isoperator(char t)
{
  int i;
  
   for (i=0;i<5;i++)
    if ( t == operators[i] )
     return 1;
    
  return 0;
}

// aliases replacer routine
void aliasesreplacer(char formula[])
{
  int i, pt=0;
  char tformula[MAXCOMMAND], textract[MAXCOMMAND];
  strcpy(tformula, formula);
    
    for (i=0;i<aliases.size();i++) {
     if ((pt=findsimple(words[separatewords(tformula, 0)].formulaWord, aliases[i].aliasPrototype))) {
      extracttextpart(tformula, textract, pt-1, strlen(aliases[i].aliasPrototype));
      inserttextpart(tformula, aliases[i].aliasText, pt-1);
     }
    }

   strcpy(formula, tformula);
}

// aliases adder and remover
int aliasesadderremover(char name[], char text[], int flag) // 0 add, 1 remove
{
  int i;
  vector<Alias>::iterator p=aliases.begin();
  
   if (!flag) {
    // does name or text contain numbers ?
    for (i=0;i<strlen(name);i++)
     if (isdigit(name[i]))
      return 0;
    for (i=0;i<strlen(text);i++)
     if (isdigit(text[i]))
      return 0;
    // look for already defined name
    if (isalreadydefined(name))
   return 0; }
  
   // find alias prototype
   for (i=0;i<aliases.size();i++)
    if (!strcmp(aliases[i].aliasPrototype, name))
     break;
   
   // depending on flag, return 0 if incompatible operation
   if (flag && i==aliases.size())
    return 0;
   if (!flag && i<aliases.size())
    return 0;
   
   // erase or add alias
   if (flag) {
    p+=i;
   aliases.erase(p); }
   else {
    Alias talias(name, text);
   aliases.push_back(talias); }
   
 return 1;
}

// list aliases
void aliaseslistview()
{
  int i;
  
   if (!aliases.size())
    cout << "no defined aliases" << endl;
   else {
    for (i=0;i<aliases.size();i++)
   cout << aliases[i].aliasPrototype << " " << aliases[i].aliasText << endl; }
}

// name is already defined
int isalreadydefined(char *name)
{
  int i;
  
   // check for identical variable or function or alias
   for (i=0;i<variables.size();i++)
    if (!strcmp(name, variables[i].variableName) || findsimple(variables[i].variableName, name) || findsimple(name, variables[i].variableName))
     return 1;
   for (i=0;i<functions.size();i++)
    if (!strcmp(name, functions[i].functionName) || findsimple(functions[i].functionName, name) || findsimple(name, functions[i].functionName))
     return 1;
   for (i=0;i<aliases.size();i++)
    if (!strcmp(name, aliases[i].aliasPrototype) || findsimple(aliases[i].aliasPrototype, name) || findsimple(name, aliases[i].aliasPrototype))
     return 1;
   // check for nonallowed nametypes
   for (i=0;i<32;i++) // 31 nonallowed nametypes sofar
    if (!strcmp(name, nonallowednametypes[i]))
     return 1;
    
 return 0;
}

// string functions parser
int stringsystemfunctionparser(int function_id, char tcommand[])
{
  int i, n, pos, endpos, parameters[2];
  char tparameter[MAXNAME+1], ttext[MAXSTRING], tttext[MAXSTRING];
  
   // count parentheses
   if ((i=countparentheses(tcommand)))
    return i;
  
   // trick algorithm below with ',' when system function which has no text
   if (functions[function_id].internalflag==3) {
    n=0;
    while (tcommand[n]!='(')
     ++n;
    tcommand[n]=',';
   }
   // go back from end of tcommand
   pos=strlen(tcommand);
   n=functions[function_id].functionParameters;
   while (n && pos)
    if (tcommand[pos--]==',')
     --n;
   if (!pos || n)
    return -1;
   pos+=2; // now pos is at first numerical parameter or space
   if (!functions[function_id].functionParameters)
    pos=strlen(tcommand);
   endpos=pos-1; // use later
   // read parameters
   for (i=0;i<functions[function_id].functionParameters;n=0, i++) {
    while (tcommand[pos]!=',' && tcommand[pos]!=')')
     tparameter[n++]=tcommand[pos++];
    tparameter[n]='\0';
    searchandsubstitutevariables(tparameter);
    ++pos;
    parameters[i]=atoi(tparameter); 
    if (!parameters[i])
     return 1;
   }
   
    // get string to apply command on
    if (functions[function_id].internalflag!=3) {
     i=0; n=0;
     while (tcommand[i++]!='(');
     for (;i<endpos;i++)
      ttext[n++]=tcommand[i];
     ttext[n]='\0';
    }
   
    // substitute variables
    searchandsubstitutevariables(ttext);
    // string functions
    n=0;
    if (!strcmp(functions[function_id].functionName, "mid$")) {
     for (i=parameters[0]-1;i<parameters[0]+parameters[1]-1;i++)
      tttext[n++]=ttext[i];
     tttext[n]='\0';
    }
    if (!strcmp(functions[function_id].functionName, "left$")) {
     for (i=0;i<parameters[0];i++)
      tttext[n++]=ttext[i];
     tttext[n]='\0';
    }
    if (!strcmp(functions[function_id].functionName, "right$")) {
     for (i=strlen(ttext)-parameters[0];i<strlen(ttext);i++)
      tttext[n++]=ttext[i];
     tttext[n]='\0';
    }
    if (!strcmp(functions[function_id].functionName, "toupper$")) {
     for (i=0;i<strlen(ttext);i++)
      tttext[n++]=toupper(ttext[i]);
     tttext[n]='\0';
    }
    if (!strcmp(functions[function_id].functionName, "tolower$")) {
     for (i=0;i<strlen(ttext);i++)
      tttext[n++]=tolower(ttext[i]);
     tttext[n]='\0';
    }
    // system functions
    if (!strcmp(functions[function_id].functionName, "toxy"))
     gotoxy((int) parameters[0], (int) parameters[1]);
    if (!strcmp(functions[function_id].functionName, "color"))
     coutf((int) parameters[0], (int) parameters[1]);
    // empty tcommand if system function
    if (functions[function_id].internalflag==3)
     strcpy(tcommand, "");
    else
     // available char string to transfer to calling routine is tcommand
     strcpy(tcommand, tttext);
        
 return 0;
}

// isstring variable
int isstringvariable(char *tvariable)
{
  if ( tvariable[strlen(tvariable)-1] == '$' )
   return 1;
    
 return 0;
}

// is rem line
int isremline(const char formula[])
{
  int i=0;
  
   while ( isspace(formula[i]) )
    ++i;
   if ( formula[i] == ';' )
    return 1;
    
 return 0;
}

// extract formula past word number
char* extractfromformula(char formula[], int word_id)
{
  int n, i=0;
  static char *extract=new char[MAXCOMMAND];
   
   for (n=0;n<word_id;n++)
    i+=strlen(words[separatewords(formula, n)].formulaWord);
   i+=word_id; // spaces
   
   n=0;
   for (;i<strlen(formula);i++)
    extract[n++]=formula[i];
   extract[n]='\0';

 return extract;
}

// count parentheses in string
int countparentheses(char *tstring)
{
  openparentheses=0;
  
   while (*tstring) {
    if (*tstring=='(')
     --openparentheses;
    if (*tstring==')')
     ++openparentheses;
    ++tstring;
   }
    
  return openparentheses;
}

// ismatrixvariable
int ismatrixvariable(char name[])
{
  int i;
  
   for (i=0;i<variables.size();i++)
    if (!strcmp(name, variables[i].variableName))
     if (variables[i].mSize.size())
      return 1;
    
 return 0;
}

// define/recall matrix variable
int definesetrecallmatrixvariable(char name[], char sizeparameters[], int flag) // 0 define, 1 recall, 2 set matrix point
{
  int i, n, l, parametersize=0, matchedparameters, matchvariable=-1, mpos=0, parameters[MAXMATRIX];
  double dp;
  char textract[MAXNAME+1];
  
   parseformulaforfunctions(sizeparameters); // necessary
   while (mpos<strlen(sizeparameters)) {
    i=0;
    while (sizeparameters[mpos]!='x' && sizeparameters[mpos]!=',' && mpos<strlen(sizeparameters))
     textract[i++]=sizeparameters[mpos++];
    textract[i]='\0';
    if (!(dp=runcalculationseries(textract)))
     return -2;
    if (!(parameters[parametersize]=(int) dp) || parameters[parametersize]>MAXMATRIXPOINTS || parametersize>MAXMATRIX-1)
     return -2;
    ++parametersize;
    ++mpos;
   }
    
    if ( !flag ) {
     if ((i=definevariable(name, emptystring))<1)
      return i;
     for (i=0;i<parametersize;i++)
      variables[variables.size()-1].mSize.push_back(parameters[i]);
    }

    else {
     for (n=0;n<variables.size();n++)
      if (!strcmp(name, variables[n].variableName))
       break;
     if (n==variables.size())
      return 0;
     if (flag==1)
      variables[n].variableValue=0;
     if (parametersize!=variables[n].mSize.size())
      return 0;
     for (i=0;i<parametersize;i++) {
      --parameters[i];
      if (parameters[i]<0 || parameters[i]>variables[n].mSize[i]-1)
      return 0;
     }
     for (i=0;i<variables[n].mValue.size();i++) {
       matchedparameters=0;
      for (l=0;l<variables[n].mSize.size();l++) {
        if (variables[n].mValue[i].pt[l]==parameters[l])
         ++matchedparameters; }
      if (matchedparameters==variables[n].mSize.size()) {
       matchvariable=i;
       break;
      }
     }

     if (matchvariable>-1) {
      if (flag==1)
       variables[n].variableValue=variables[n].mValue[matchvariable].matrixValue;
      else
       variables[n].mValue[matchvariable].matrixValue=variables[n].variableValue;
     }
     else if (flag==2) {
      MatrixValue tmvalue;
      tmvalue.matrixValue=variables[n].variableValue;
      for (i=0;i<variables[n].mSize.size();i++)
       tmvalue.pt.push_back(parameters[i]);
      variables[n].mValue.push_back(tmvalue);
     }
    }
     
 return 1;
}

// read matrix parameters and rewrite 
int readmatrixvariableparameters(char formula[])
{
  int mpos=0, n=0, parameters[MAXMATRIX], i;
  double dmvalue;
  char matrixvariable[MAXNAME+1], tformula[MAXNAME+1], textract[MAXNAME+1];
  
   while (formula[mpos]!='[' && mpos<strlen(formula)) {
    if (formula[mpos]!='[')
     matrixvariable[n++]=formula[mpos]; 
   ++mpos; }
   matrixvariable[n]='\0';
   n=0;
   if (mpos==strlen(formula))
    return -2;
   ++mpos;
   while (formula[mpos]!=']' && mpos<strlen(formula)) {
    if (mpos!=']' && mpos<strlen(formula))
     tformula[n++]=formula[mpos];
   ++mpos; }
   tformula[n]='\0';
   if (mpos==strlen(tformula))
    return -2;

   definesetrecallmatrixvariable(matrixvariable, tformula, 1);
  
 return 1;
}


